#ifdef _DEBUG
#include "DebugHelpers.h"
#endif
#include "amule.h"
#include "updownclient.h"
#include "Friend.h"
#include "ClientList.h"
#include "OtherFunctions.h"
#include "PartFile.h"
#include "ListenSocket.h"
#include "Friend.h"
#include <zlib.h>
#include "Packet.h"
#include "OPCodes.h"
#include "SafeFile.h"
#include "Preferences.h"
#include "Statistics.h"
#include "ClientCredits.h"
#include "IPFilter.h"
#include "UploadQueue.h"
#include "DownloadQueue.h"
#include "SearchList.h"
#include "SharedFileList.h"
#ifdef __USE_KAD__
#include "Kademlia/Kademlia/Kademlia.h"
#include "Kademlia/Net/KademliaUDPListener.h"
#endif
#if !defined(_CONSOLE) && !defined(AMULE_DAEMON)
#include "ServerWnd.h"
#include "TransferWnd.h"
#include "ChatWnd.h"
#endif

#include <cmath>

#include "AdunanzA.h"
#include "RemoteSettings.h"

// Inclusi per il parsing sul file per Updater
#include <stdio.h>
#include <stdlib.h>
#define thePrefs (*(theApp.glob_prefs))
// Funzione molto semplice che ritorna un 
// valore booleano vero se l'ip passatogli in input 
// appartiene alla rete FastWeb italiana.
/*
MAN Milano........................   1.0.0.0/8
MAN Milano hinterland Nord........   2.0.0.0/8
MAN Genova........................   5.0.0.0/8
MAN Veneto........................  11.0.0.0/8
MAN Milano hinterland Sud.........  14.0.0.0/8
MAN Ancona........................  21.0.0.0/8
MAN Grosseto......................  22.0.0.0/8
MAN Roma..........................  23.0.0.0/8
MAN Toscana.......................  27.0.0.0/8
MAN Triveneto.....................  29.0.0.0/8
MAN Bari..........................  31.0.0.0/8
MAN Sicilia.......................  36.0.0.0/8
MAN Bologna.......................  37.0.0.0/8
MAN Napoli........................  39.0.0.0/8
MAN Torino........................  41.0.0.0/8
MAN Reggio Emilia.................  42.0.0.0/8

Routers & OAM.....................  10.0.0.0/8
Organization, Administration & Management (ad. es.: Server interni ntp, voip, billing etc. etc.)
*/

bool AduIsFastWebLANIP(uint32 ip)
{
	switch (ip & 0xff) {
		case MAN_ROMA:
		case MAN_MILANO:
		case MAN_MILANO_H_NORD:
		case MAN_MILANO_H_SUD:
		case MAN_GENOVA:
		case MAN_TRIVENETO:
		case MAN_BARI:
		case MAN_BOLOGNA:
		case MAN_NAPOLI:
		case MAN_TORINO:
		case MAN_REGGIO_EMILIA:
		case MAN_TOSCANA:
		case MAN_VENETO:
		case MAN_GROSSETO:
		case MAN_SICILIA:
    case MAN_ANCONA:
		case FASTWEB_ROUTERS:
			return true;
		default:
			return false;
	}
}

void CUpDownClient::SetClientFWCity(void)
{
	switch (m_dwUserIP & 0x000000ff) {
		case MAN_ROMA:
			m_FWCity = wxT("RM");
			break;
		case MAN_MILANO:
			m_FWCity = wxT("MI");
			break;
		case MAN_MILANO_H_NORD:
			m_FWCity = wxT("MI H. Nord");
			break;
		case MAN_MILANO_H_SUD:
			m_FWCity = wxT("MI H. Sud");
			break;
		case MAN_GENOVA:
			m_FWCity = wxT("GE");
			break;
		case MAN_TRIVENETO:
			m_FWCity = wxT("Triveneto");
			break;
		case MAN_BARI:
			m_FWCity = wxT("BA");
			break;
		case MAN_BOLOGNA:
			m_FWCity = wxT("BO");
			break;
		case MAN_NAPOLI:
			m_FWCity = wxT("NA");
			break;
		case MAN_TORINO:
			m_FWCity = wxT("TO");
			break;
		case MAN_REGGIO_EMILIA:
			m_FWCity = wxT("RE");
			break;
		case MAN_TOSCANA:
			m_FWCity = wxT("Toscana");
			break;
		case MAN_VENETO:
			m_FWCity = wxT("Veneto");
			break;
		case MAN_GROSSETO:
			m_FWCity = wxT("GR");
			break;
		case MAN_SICILIA:
			m_FWCity = wxT("Sicilia");
			break;
		case MAN_ANCONA:
			m_FWCity = wxT("AN");
			break;
		case FASTWEB_ROUTERS:
			m_FWCity = wxT("OAM");
			break;
		default:
			m_FWCity = wxT("");
			break;
	}
}



bool CUpDownClient::IsAduFiber() const
{
	return AduIsFiber(m_dwUserIP);
}

bool AduIsFiber (uint32 ip) {
  bool	is128 = (((ip >> 15) & 1) == (uint32)0);

	return AduIsFastWebLANIP(ip) & is128;
}

bool AduIsFastWebIP(uint32 ip)
{
	if (AduIsFastWebLANIP(ip)) 
		return true;
	
	// match rapido: (ip & mask) == netid

	// netid 213.140.0.0/19			- netid (hex) 0x00008cd5 / mask 0x00e0ffff
	if ((ip & 0x00e0ffff) == 0x00008cd5 ||
	// netid 213.156.32.0/19		- netid (hex) 0x00209cd5 / mask 0x00e0ffff
			(ip & 0x00e0ffff) == 0x00209cd5 ||
	// netid 62.101.64.0/18			- netid (hex) 0x0040653e / mask 0x00c0ffff
			(ip & 0x00c0ffff) == 0x0040653e ||
	// netid 81.208.0.0/17			-	netid (hex) 0x0000d051 / mask 0x0080ffff
			(ip & 0x0080ffff) == 0x0000d051 ||
	// netid 83.103.0.0/17			- netid (hex) 0x00006753 / mask 0x0080ffff
			(ip & 0x0080ffff) == 0x00006753 ||
	// netid 85.18.0.0/16       - netid (hex) 0x00001255 / mask 0x0000ffff
			(ip & 0x0000ffff) == 0x00001255 ||
	// netid 89.96.0.0/15       - netid (hex) 0x00006059 / mask 0x0000f7ff
			(ip & 0x0000f7ff) == 0x00006059) return true;


	return false;
}

#include "Logger.h"
#include <common/Format.h>
#include "NetworkFunctions.h"

bool     AduIsValidKaduAddress( uint32 host )
{
	host = wxUINT32_SWAP_ALWAYS( host );
	//AddDebugLogLineM(false, logClientKadUDP, CFormat(wxT("Checking %s for kADU validity")) % Uint32toStringIP(host));

	// TODO: Add Banlist support here

	return AduIsFastWebIP( host );
}

// Metodo dell'oggetto CUpDownClient
// Vedere il file 
// Questo metodo ritorna 
// ADUNANZA_ICON_NONE	Se il client non e' adunanza
// ADUNANZA_ICON_FW		Se il client non e' adunanza ma e' sulla rete FastWeb Italia
// ADUNANZA_ICON_ADU	Se il client e' AdunanzA ed e' sulla rete FastWeb Italia
uint32 CUpDownClient::GetClientAduType(void) const
{
	if (AduIsFastWebIP(m_dwUserIP))
	{
		if (GetIsAduSoftware())
			return ADUNANZA_ICON_ADU;
		else 
			return ADUNANZA_ICON_FW;
	}
	return ADUNANZA_ICON_NONE;
}

bool CUpDownClient::IsAduClient() const
{
	return GetClientAduType() == ADUNANZA_ICON_ADU;
}

float calcolaStima(float avail, uint32& firstPublish, uint32 publishInterval, uint32 pubkRTK, bool sameIP, uint32 now) {
	int kRTK = theApp.rm->kadRepublishTimeK;

	// Using this increment style, we can guess also if clients use different publish rates.
	float inc = (float)pubkRTK / kRTK;
	
	// Only old clients, with republishtimek == 24hrs don't publish their republishtimek.
	if (inc == 0.f)
  	inc = HR2S(24) / kRTK;	

	if (publishInterval > kRTK)
		avail=0.f,firstPublish=now;	

	while ((now-firstPublish) > kRTK && avail > 2.f)
		avail--,firstPublish+=(uint32)(kRTK/avail);

	// Some client seem to send duplicate publish to already contacted clients.
	// In 90% of the times we avoid this here.
  if (!sameIP)
    avail+=inc;

  return avail;
}

float normalizzaStima( float avail, uint32 from, uint32 to ) {
	// Parte necessaria per il debug
  // calcolo il coefficiente di normalizzazione
	int kRTK = theApp.rm->kadRepublishTimeK;

 	float norm_factor = 1.f;

	if (avail > theApp.rm->kadFreshGuess_NoNorm) {
		norm_factor = (to - from)/kRTK;

	if ( norm_factor > 1-1/avail )
		norm_factor = 1.f;

	norm_factor = ( norm_factor < theApp.rm->kadFreshGuess_Tol ? theApp.rm->kadFreshGuess_Tol : norm_factor );

	// We have a good number of publishes, but not high.
	// Low normalization is still suggested.
	if ( avail < theApp.rm->kadFreshGuess_LowNorm && norm_factor > avail / theApp.rm->kadFreshGuess_LowNorm )
		norm_factor /= powf( avail/theApp.rm->kadFreshGuess_LowNorm, 0.5f );
	}

  if ( avail > 2.f )
    avail /= powf( norm_factor, theApp.rm->kadFreshGuess_Weight );

  return avail;
}

#include "Statistics.h"


// Decido come selezionare un client in base alla
// banda realmente data in upload
uint32 AduGetTypeBand()
{
	uint64 sessionSentBytes = theApp.statistics->GetSessionSentBytes();
	uint64 sessionRecvBytes = theApp.statistics->GetSessionReceivedBytes();

	uint64 aduSessionSentBytes = theApp.statistics->GetAduSessionSentBytes();
	uint64 aduSessionRecvBytes = theApp.statistics->GetAduSessionReceivedBytes();

	uint64 sentToExternal   = sessionSentBytes - aduSessionSentBytes;
	uint64 recvFromExternal = sessionRecvBytes - aduSessionRecvBytes;

	if ( sentToExternal < recvFromExternal )
		return ADUNANZA_EXTERN;
	else 
		return ADUNANZA_FASTWEB;
}

// lista di servers solo interna
typedef struct	_servIP_Port
{
	uint32	ip;
	uint16	port;
} servIP_Port, *LPservIP_Port;

static LPservIP_Port	_aduServers = NULL;
static int				_aduServersNum = 0;


// Ci dice se connesso ad un server adunanza o meno
bool IsAduServer(uint32 CurServer) {

	if (_aduServers && _aduServersNum)
	{
		for(int i = 0; i < _aduServersNum; i++)
		{
			if (CurServer == _aduServers[i].ip) return true;
		}
	}
	return false;
}

#ifndef CLIENT_GUI
int AduQ(void) {
  return theApp.uploadqueue->GetAdunanzAUserCount();
}

bool AduDebt(void) {
	uint64 sentToExternal   = theApp.statistics->GetSessionSentBytes() - theApp.statistics->GetAduSessionSentBytes();
	uint64 recvFromExternal = theApp.statistics->GetSessionReceivedBytes() - theApp.statistics->GetAduSessionReceivedBytes();

	if ( sentToExternal < recvFromExternal )
	  return true;
  else 
    return false;
}
int SetSlots(void) {
	float maxUp = (float)thePrefs.GetMaxUpload();
 	float kBpsUpPerClient = (float)thePrefs.GetSlotAllocation();
	float kBpsUp = (float)theStats::GetUploadRate() / 1024.0f;
  int slots;


	if (thePrefs.GetMaxUpload() == UNLIMITED)
		slots = (int)( kBpsUp / kBpsUpPerClient + 2 );
	else
		slots = (int)floor( maxUp / kBpsUpPerClient + 0.5f );

  return std::max(MIN_UP_CLIENTS_ALLOWED,slots);
}

int TotSlots(void) {
	if (AduDebt())
		return (int)(SetSlots() * 1.5f);
	else
		return SetSlots();
}

int ExtSlots(void) {
	if (AduDebt())
		return (int)(TotSlots() * 0.5f);
	else
		return TotSlots();
}

uint32 AduNextClient(int fix) {
	int totslots = TotSlots(), totcount = theApp.uploadqueue->GetUploadCount() - fix;

	if (totcount < totslots) {
    if (!AduQ())
      return ADUNANZA_ANY;

		int aducount = theApp.uploadqueue->GetAduUploadCount();
		int extcount = totcount - aducount;

		if (AduDebt()) {
			int extslots = ExtSlots() - fix;
			int aduslots = totslots - extslots;

		  if (extslots > extcount && aduslots > aducount)
				return ADUNANZA_ANY;
			else if (extslots > extcount)
				return ADUNANZA_EXTERN;
			else if (aduslots > aducount)
				return ADUNANZA_FASTWEB;
			else
				return ADUNANZA_NONE; // non dovrebbe poter succedere

		} else {
			return ADUNANZA_FASTWEB;
		}
	} else
		return ADUNANZA_NONE;
}

uint32 AduGetMaxUploadSlots(void)
{
	return TotSlots();
}

bool AduMaxTrans(void) {
  uint32 now = (uint32)time(NULL);
  if (AduQ() && AduNextClient(1) == ADUNANZA_FASTWEB && now > theApp.uploadqueue->lastmax + 45 ) {
    theApp.uploadqueue->lastmax = now;
    return true;
  } else
    return false;
}
#endif
