#include "RemoteSettings.h"
#include "OPCodes.h"
#include "InternalEvents.h"                // Needed for CMuleInternalEvent
#include "OtherFunctions.h" // needed for GetConfigDir()
#include "HTTPDownload.h"   // needed for CHTTPDownloader
#include <wx/filefn.h>
#include <common/StringFunctions.h>
#include "AdunanzA.h"

#define RSEXPIRETIME HR2S(96)

#define RSRECHECKTIME HR2S(6)

#define FAILCHECKSTATIC 3

#define ADURM_TMPFILE "_tmp_adunanza.conf"

CRemoteSettings::CRemoteSettings() {
	kadRepublishTimeK = KADEMLIAREPUBLISHTIMEK;
	kadRepublishTimeS = KADEMLIAREPUBLISHTIMES;
	kadRepublishTimeN = KADEMLIAREPUBLISHTIMEN;

	kadTotalStoreKey = KADEMLIATOTALSTOREKEY;
	kadTotalStoreSrc = KADEMLIATOTALSTORESRC;
	kadTotalStoreNotes = KADEMLIATOTALSTORENOTES;

	kadTotalSearchFile = KADEMLIATOTALFILE;

	kadMaxSrcFile = KADEMLIAMAXSOUCEPERFILE;
	kadMaxNotFile = KADEMLIAMAXNOTESPERFILE;

	kadFreshGuess_Tol = KADEMLIAFRESHGUESS_TOLERANCE;
	kadFreshGuess_Weight = KADEMLIAFRESHGUESS_WEIGHT;
	kadFreshGuess_NoNorm = KADEMLIAFRESHGUESS_NONORMSOURCES;
	kadFreshGuess_LowNorm = KADEMLIAFRESHGUESS_LOWNORMSOURCES;

	maxSrc = MAX_SOURCES_FILE_SOFT;
	maxSrcUdp = MAX_SOURCES_FILE_UDP;

	kadFindValue = KADEMLIA_FIND_VALUE;
	kadStore = KADEMLIA_STORE;
	kadFindNode = KADEMLIA_FIND_NODE;

	kadReaskTime = KADEMLIAREASKTIME;
	kadPubTime = KADEMLIAPUBLISHTIME;

	kadReaskIncs = 7;

	nextUpdate = 0;
	failCount = 0;
	UpdateURL = wxT(ADURM_URL);
	expireTime = RSEXPIRETIME;
	recheckTime = RSRECHECKTIME;

	kadOpcode = OP_KADEMLIAHEADER;
	kadZOpcode = OP_KADEMLIAPACKEDPROT;

  mVer = 0;

	ConfigDir = GetConfigDir();

	cfg = new wxFileConfig(wxEmptyString, wxEmptyString, ConfigDir + wxT("adunanza.conf") );

	wxString nU;
	cfg->Read(wxT("/Adunanza/nextUpdate") , &nU, wxT("0"));
	nextUpdate = strtoul(unicode2char(nU.c_str()),NULL,10);
	cfg->Read(wxT("/Adunanza/failCount") , &failCount, failCount);

	//printf("---------------- On init -----------------\n");
	//PrintAllValues();

	ReadSettings(cfg);

	if ( nextUpdate > time(NULL) + (time_t)expireTime )
		nextUpdate = time(NULL) + (time_t)expireTime ;

	//printf("---------------- After reading -----------------\n");
	//PrintAllValues();
}

CRemoteSettings::~CRemoteSettings() {
	//printf("---------------- On close -----------------\n");
	//PrintAllValues();

	SaveSettings();
	delete cfg;
}

void CRemoteSettings::PrintAllValues() {
#if ADU_BETA_MAJ > 0
	printf("kadRepublishTimeK: %ld\n",kadRepublishTimeK);
	printf("kadRepublishTimeS: %ld\n",kadRepublishTimeS);
	printf("kadRepublishTimeN: %ld\n",kadRepublishTimeN);

	printf("kadTotalStoreKey: %ld\n",kadTotalStoreKey);
	printf("kadTotalStoreSrc: %ld\n",kadTotalStoreSrc);
	printf("kadTotalStoreNotes: %ld\n",kadTotalStoreNotes);

	printf("kadTotalSearchFile: %ld\n",kadTotalSearchFile);

	printf("kadMaxSrcFile: %ld\n",kadMaxSrcFile);
	printf("kadMaxNotFile: %ld\n",kadMaxNotFile);

	printf("kadFreshGuess_Tol: %.2f\n",kadFreshGuess_Tol);
	printf("kadFreshGuess_Weight: %.2f\n",kadFreshGuess_Weight);
	printf("kadFreshGuess_LowNorm: %ld\n",kadFreshGuess_LowNorm);
	printf("kadFreshGuess_NoNorm: %ld\n",kadFreshGuess_NoNorm);

	printf("maxSrc: %ld\n",maxSrc);
	printf("maxSrcUdp: %ld\n",maxSrcUdp);

	printf("kadFindValue: %ld\n",kadFindValue);
	printf("kadStore: %ld\n",kadStore);
	printf("kadFindNode: %ld\n",kadFindNode);

	printf("kadReaskTime: %ld\n",kadReaskTime);
	printf("kadReaskIncs: %ld\n",kadReaskIncs);
	printf("kadPubTime: %ld\n",kadPubTime);
	printf("mVer: %ld\n",mVer);

	printf("opcodes Norm: 0x%02x Zip: 0x%02x\n", kadOpcode, kadZOpcode);
	printf("nextUpdate in %lu seconds\n",nextUpdate-time(NULL));
	printf("expireTime: %d recheckTime: %d failCount: %d\n",expireTime,recheckTime,failCount);
	printf("updating from %s\n",(const char*)unicode2char(UpdateURL));
#endif
}

void CRemoteSettings::SaveToFile() {
	SaveSettings();
	delete cfg;
	cfg = new wxFileConfig(wxEmptyString, wxEmptyString, ConfigDir + wxT("adunanza.conf") );
}

void CRemoteSettings::ReadSettings(wxConfigBase *docfg) {
	docfg->Read(wxT("/Adunanza/expireTime") , &expireTime, expireTime);
	docfg->Read(wxT("/Adunanza/recheckTime") , &recheckTime, recheckTime);
	docfg->Read(wxT("/Adunanza/UpdateURL") , &UpdateURL, UpdateURL);

	if (expireTime > RSEXPIRETIME)
		expireTime = RSEXPIRETIME;

	if (recheckTime > RSRECHECKTIME)
		recheckTime = RSRECHECKTIME;

	docfg->Read(wxT("/Adunanza/kadRepublishTimeK") , &kadRepublishTimeK, kadRepublishTimeK);
	docfg->Read(wxT("/Adunanza/kadRepublishTimeS") , &kadRepublishTimeS, kadRepublishTimeS);
	docfg->Read(wxT("/Adunanza/kadRepublishTimeN") , &kadRepublishTimeN, kadRepublishTimeN);

	docfg->Read(wxT("/Adunanza/kadTotalStoreKey") , &kadTotalStoreKey, kadTotalStoreKey);
	docfg->Read(wxT("/Adunanza/kadTotalStoreSrc") , &kadTotalStoreSrc, kadTotalStoreSrc);
	docfg->Read(wxT("/Adunanza/kadTotalStoreNotes") , &kadTotalStoreNotes, kadTotalStoreNotes);

	docfg->Read(wxT("/Adunanza/kadTotalSearchFile") , &kadTotalSearchFile, kadTotalSearchFile);

	docfg->Read(wxT("/Adunanza/kadMaxSrcFile") , &kadMaxSrcFile, kadMaxSrcFile);
	docfg->Read(wxT("/Adunanza/kadMaxNotFile") , &kadMaxNotFile, kadMaxNotFile);

	kadFreshGuess_Tol = docfg->Read(wxT("/Adunanza/kadFreshGuess_Tol") , (long)(kadFreshGuess_Tol * 10000)) / 10000.f;
	kadFreshGuess_Weight = docfg->Read(wxT("/Adunanza/kadFreshGuess_Weight") , (long)(kadFreshGuess_Weight * 10000)) / 10000.f;
	docfg->Read(wxT("/Adunanza/kadFreshGuess_LowNorm") , &kadFreshGuess_LowNorm, kadFreshGuess_LowNorm);
	docfg->Read(wxT("/Adunanza/kadFreshGuess_NoNorm") , &kadFreshGuess_NoNorm, kadFreshGuess_NoNorm);

	docfg->Read(wxT("/Adunanza/maxSrc") , &maxSrc, maxSrc);
	docfg->Read(wxT("/Adunanza/maxSrcUdp") , &maxSrcUdp, maxSrcUdp);

	docfg->Read(wxT("/Adunanza/kadFindValue") , &kadFindValue, kadFindValue);
	docfg->Read(wxT("/Adunanza/kadStore") , &kadStore, kadStore);
	docfg->Read(wxT("/Adunanza/kadFindNode") , &kadFindNode, kadFindNode);

	docfg->Read(wxT("/Adunanza/kadReaskTime") , &kadReaskTime, kadReaskTime);
	docfg->Read(wxT("/Adunanza/kadReaskIncs") , &kadReaskIncs, kadReaskIncs);
	docfg->Read(wxT("/Adunanza/kadPubTime") , &kadPubTime, kadPubTime);
	docfg->Read(wxT("/Adunanza/mVer") , &mVer, mVer);

	kadOpcode = (unsigned char)docfg->Read(wxT("/Adunanza/kadOpcode") , (long)kadOpcode);
	kadZOpcode = (unsigned char)docfg->Read(wxT("/Adunanza/kadZOpcode") , (long)kadZOpcode);
	
}

void CRemoteSettings::SaveSettings() {
	cfg->Write(wxT("/Adunanza/nextUpdate") , wxString::Format(wxT("%ul"),nextUpdate));
	cfg->Write(wxT("/Adunanza/failCount") , failCount);
	cfg->Write(wxT("/Adunanza/UpdateURL") , UpdateURL);
	cfg->Write(wxT("/Adunanza/expireTime") , expireTime);
	cfg->Write(wxT("/Adunanza/recheckTime") , recheckTime);

	cfg->Write(wxT("/Adunanza/kadRepublishTimeK") , kadRepublishTimeK);
	cfg->Write(wxT("/Adunanza/kadRepublishTimeS") , kadRepublishTimeS);
	cfg->Write(wxT("/Adunanza/kadRepublishTimeN") , kadRepublishTimeN);

	cfg->Write(wxT("/Adunanza/kadTotalStoreKey") , kadTotalStoreKey);
	cfg->Write(wxT("/Adunanza/kadTotalStoreSrc") , kadTotalStoreSrc);
	cfg->Write(wxT("/Adunanza/kadTotalStoreNotes") , kadTotalStoreNotes);

	cfg->Write(wxT("/Adunanza/kadTotalSearchFile") , kadTotalSearchFile);

	cfg->Write(wxT("/Adunanza/kadMaxSrcFile") , kadMaxSrcFile);
	cfg->Write(wxT("/Adunanza/kadMaxNotFile") , kadMaxNotFile);

	cfg->Write(wxT("/Adunanza/kadFreshGuess_Tol") , (long)(kadFreshGuess_Tol * 10000));
	cfg->Write(wxT("/Adunanza/kadFreshGuess_Weight") , (long)(kadFreshGuess_Weight * 10000));
	cfg->Write(wxT("/Adunanza/kadFreshGuess_LowNorm") , kadFreshGuess_LowNorm);
	cfg->Write(wxT("/Adunanza/kadFreshGuess_NoNorm") , kadFreshGuess_NoNorm);

	cfg->Write(wxT("/Adunanza/maxSrc") , maxSrc);
	cfg->Write(wxT("/Adunanza/maxSrcUdp") , maxSrcUdp);

	cfg->Write(wxT("/Adunanza/kadFindValue") , kadFindValue);
	cfg->Write(wxT("/Adunanza/kadStore") , kadStore);
	cfg->Write(wxT("/Adunanza/kadFindNode") , kadFindNode);

	cfg->Write(wxT("/Adunanza/kadReaskTime") , kadReaskTime);
	cfg->Write(wxT("/Adunanza/kadReaskIncs") , kadReaskIncs);
	cfg->Write(wxT("/Adunanza/kadPubTime") , kadPubTime);
	cfg->Write(wxT("/Adunanza/mVer") , mVer);

	cfg->Write(wxT("/Adunanza/kadOpcode"), (long)kadOpcode);
	cfg->Write(wxT("/Adunanza/kadZOpcode"), (long)kadZOpcode);
}

void CRemoteSettings::CheckUpdate() {
	if (time(NULL) < (time_t)nextUpdate)
		return;

	//printf("---------------- Before update -----------------\n");
	//PrintAllValues();

	wxString strURL(UpdateURL);
	wxString tmpFile(ConfigDir + wxT(ADURM_TMPFILE));
	if (wxFileExists(tmpFile))
		wxRemoveFile(tmpFile);

	CHTTPDownloadThread *downloader = new CHTTPDownloadThread(strURL, tmpFile, HTTP_AdunanzaConf);
	downloader->Create();
	downloader->Run();
}

void CRemoteSettings::FileDownloadedCallback(int result) {
	//printf("---------------- Before update (2) ----------------- result: %d\n",result);
	//PrintAllValues();

	if (result == -1) {
		double scaled = rand() / (RAND_MAX + 1.0);
		nextUpdate = time(NULL) + (time_t)recheckTime + (time_t)((recheckTime + 1) * scaled);
		failCount++;
		if (failCount > FAILCHECKSTATIC) {
			UpdateURL = wxT(ADURM_URL);
			SaveSettings();
		}
		return;
	}

	wxString tmpFile(ConfigDir + wxT(ADURM_TMPFILE));
	if (!wxFileExists(tmpFile))
		return;

	wxFileConfig *tmpcfg = new wxFileConfig(wxEmptyString, wxEmptyString, tmpFile);

	ReadSettings(tmpcfg);

	delete tmpcfg;

	wxRemoveFile(tmpFile);

	SaveToFile();

	nextUpdate = time(NULL) + (time_t)expireTime;
	failCount = 0;

	printf("---------------- After update -----------------\n");
	PrintAllValues();
}
